@@ -3,7 +3,7 @@
 /* header file for S-Lang internal structures that users do not (should not)
    need.  Use slang.h for that purpose. */
 /*
-Copyright (C) 2004, 2005 John E. Davis
+Copyright (C) 2004-2008 John E. Davis
 
 This file is part of the S-Lang Library.
 
@@ -91,8 +91,8 @@ typedef enum
    SLANG_BC_LVARIABLE_APUT	= 0x29,
    SLANG_BC_LOBJPTR		= 0x2A,
    SLANG_BC_GOBJPTR		= 0x2B,
-   SLANG_BC_UNUSED_0x2C		= 0x2C,
-   SLANG_BC_UNUSED_0x2D		= 0x2D,
+   SLANG_BC_FIELD_REF		= 0x2C,
+   SLANG_BC_DEREF_FUN_CALL	= 0x2D,
    SLANG_BC_UNUSED_0x2E		= 0x2E,
    SLANG_BC_UNUSED_0x2F		= 0x2F,
    SLANG_BC_UNUSED_0x30		= 0x30,
@@ -174,7 +174,9 @@ typedef enum
    /* SLANG_BC_ARG_LVARIABLE	= 0x72, */
    SLANG_BC_UNUSED_0x72		= 0x72,
    SLANG_BC_EARG_LVARIABLE	= 0x73,
-   SLANG_BC_LINE_NUM		= 0x74,
+#define USE_BC_LINE_NUM	0	       /* used in slang.c */
+   /* SLANG_BC_LINE_NUM		= 0x74, */
+   SLANG_BC_UNUSED_0x74		= 0x74,
    SLANG_BC_BOS			= 0x75,
    SLANG_BC_EOS			= 0x76,
    SLANG_BC_UNUSED_0x77		= 0x77,
@@ -222,7 +224,7 @@ typedef enum
    SLANG_BC_LITERAL_AGET1	= 0x9F,
    SLANG_BC_LVARIABLE_APUT1	= 0xA0,
    SLANG_BC_LITERAL_APUT1	= 0xA1,
-   SLANG_BC_UNUSED_0xA2		= 0xA2,
+   SLANG_BC_LLVARIABLE_BINARY_PLUS	= 0xA2,
    SLANG_BC_UNUSED_0xA3		= 0xA3,
    SLANG_BC_UNUSED_0xA4		= 0xA4,
    SLANG_BC_UNUSED_0xA5		= 0xA5,
@@ -344,11 +346,17 @@ _pSLang_BC_Type;
 
 #define SLANG_BCST_IF		0x20
 #define SLANG_BCST_IFNOT	0x21
-#define SLANG_BCST_ELSE	0x22
+#define SLANG_BCST_ELSE		0x22
 #define SLANG_BCST_ANDELSE	0x23
 #define SLANG_BCST_ORELSE	0x24
 #define SLANG_BCST_SWITCH	0x25
 #define SLANG_BCST_NOTELSE	0x26
+#define SLANG_BCST_SC_OR	0x27
+#define SLANG_BCST_SC_AND	0x28
+
+#define SLANG_BCST_LOOP_ELSE	0x29
+#define SLANG_BCST_LOOP_THEN	0x30
+#define SLANG_BCST_COMPARE      0x31
 
 /* assignment (SLANG_BC_SET_*_LVALUE) subtypes.  The order MUST correspond
  * to the assignment token order with the ASSIGN_TOKEN as the first!
@@ -507,7 +515,7 @@ _pSL_Object_Union_Type;
 
 typedef struct _pSLang_Object_Type
 {
-   SLtype data_type;	       /* SLANG_INT_TYPE, ... */
+   SLtype o_data_type;	       /* SLANG_INT_TYPE, ... */
    _pSL_Object_Union_Type v;
 }
 SLang_Object_Type;
@@ -523,7 +531,7 @@ extern int _pSLang_pop_object_of_type (SLtype, SLang_Object_Type *, int);
 
 typedef struct
 {
-   char *name;			       /* slstring */
+   SLFUTURE_CONST char *name;			       /* slstring */
    SLang_Object_Type obj;
 }
 _pSLstruct_Field_Type;
@@ -538,28 +546,35 @@ typedef struct _pSLang_Struct_Type
 }
 _pSLang_Struct_Type;
 
-extern void _pSLstruct_delete_struct (_pSLang_Struct_Type *);
-extern int _pSLang_push_struct (_pSLang_Struct_Type *);
-extern int _pSLang_pop_struct (_pSLang_Struct_Type **);
 extern int _pSLstruct_init (void);
-/* extern int _pSLstruct_get_field (char *); */
 extern int _pSLstruct_define_struct (void);
+extern int _pSLstruct_define_struct2 (void);
 extern int _pSLstruct_define_typedef (void);
 
+extern SLang_Object_Type *_pSLstruct_get_field_value (SLang_Struct_Type *, SLCONST char *);
+extern int _pSLstruct_push_field_ref (SLFUTURE_CONST char *);
+
+int _pSLang_get_qualifiers (SLang_Struct_Type **);
+
 struct _pSLang_Ref_Type
 {
-   int is_global;
-   union
-     {
-	SLang_Name_Type *nt;
-	SLang_Object_Type *local_obj;
-     }
-   v;
+   int num_refs;
+   VOID_STAR data;
+   unsigned int sizeof_data;
+   int data_is_nametype;	       /* used for optimization */
+   int (*deref_assign)(VOID_STAR);
+   int (*deref) (VOID_STAR);
+   char *(*string)(VOID_STAR);		       /* returns a malloced string */
+   void (*destroy) (VOID_STAR);
+   int (*is_initialized) (VOID_STAR);
+   int (*uninitialize) (VOID_STAR);
 };
 
 extern int _pSLang_dereference_ref (SLang_Ref_Type *);
 extern int _pSLang_deref_assign (SLang_Ref_Type *);
-extern int _pSLang_push_ref (int, VOID_STAR);
+extern int SLang_push_ref (SLang_Ref_Type *);
+extern int _pSLang_push_nt_as_ref (SLang_Name_Type *);
+extern SLang_Ref_Type *_pSLang_new_ref (unsigned int);
 
 extern int _pSL_increment_frame_pointer (void);
 extern int _pSL_decrement_frame_pointer (void);
@@ -612,8 +627,8 @@ extern int _pSLstring_list_push (_pSLString_List_Type *, int);
 extern SLang_Array_Type *_pSLstrings_to_array (char **strs, unsigned int n);
 
 /* This function assumes that s is an slstring. */
-extern char *_pSLstring_dup_slstring (char *);
-extern int _pSLang_dup_and_push_slstring (char *);
+extern SLCONST char *_pSLstring_dup_slstring (SLCONST char *);
+extern int _pSLang_dup_and_push_slstring (SLCONST char *);
 
 
 extern int _pSLang_init_import (void);
@@ -621,7 +636,7 @@ extern int _pSLinit_exceptions (void);
 
 /* This function checks to see if the referenced object is initialized */
 extern int _pSLang_is_ref_initialized (SLang_Ref_Type *);
-extern int _pSLcheck_identifier_syntax (char *);
+extern int _pSLcheck_identifier_syntax (SLCONST char *);
 extern int _pSLang_uninitialize_ref (SLang_Ref_Type *);
 
 extern int _pSLpush_slang_obj (SLang_Object_Type *);
@@ -629,6 +644,8 @@ extern int _pSLpush_slang_obj (SLang_Object_Type *);
 extern char *_pSLexpand_escaped_char(char *, SLwchar_Type *, int *);
 extern void _pSLexpand_escaped_string (char *, char *, char *);
 
+extern int _pSLpush_dollar_string (SLFUTURE_CONST char *);
+
 /* returns a pointer to an SLstring string-- use SLang_free_slstring */
 extern char *_pSLstringize_object (SLang_Object_Type *);
 extern int _pSLdump_objects (char *, SLang_Object_Type *, unsigned int, int);
@@ -641,57 +658,60 @@ extern int _pSLang_peek_at_stack2 (SLtype *);
 struct _pSLang_NameSpace_Type
 {
    struct _pSLang_NameSpace_Type *next;
-   char *name;			       /* this is the load_type name */
-   char *namespace_name;	       /* this name is assigned by implements */
-   char *private_name;
+   SLFUTURE_CONST char *name;			       /* this is the load_type name */
+   SLFUTURE_CONST char *namespace_name;	       /* this name is assigned by implements */
+   SLFUTURE_CONST char *private_name;
    unsigned int table_size;
    SLang_Name_Type **table;
 };
-extern SLang_NameSpace_Type *_pSLns_new_namespace (char *, unsigned int);
-extern SLang_NameSpace_Type *_pSLns_allocate_namespace (char *, unsigned int);
+extern SLang_NameSpace_Type *_pSLns_new_namespace (SLFUTURE_CONST char *, unsigned int);
+extern SLang_NameSpace_Type *_pSLns_allocate_namespace (SLFUTURE_CONST char *, unsigned int);
 extern void _pSLns_deallocate_namespace (SLang_NameSpace_Type *);
-extern SLang_NameSpace_Type *_pSLns_find_namespace (char *);
-extern int _pSLns_set_namespace_name (SLang_NameSpace_Type *, char *);
-extern SLang_Array_Type *_pSLnspace_apropos (SLang_NameSpace_Type *, char *, unsigned int);
+extern SLang_NameSpace_Type *_pSLns_find_namespace (SLCONST char *);
+extern int _pSLns_set_namespace_name (SLang_NameSpace_Type *, SLFUTURE_CONST char *);
+extern SLang_Array_Type *_pSLnspace_apropos (SLang_NameSpace_Type *, SLFUTURE_CONST char *, unsigned int);
 extern void _pSLang_use_namespace_intrinsic (char *name);
-extern char *_pSLang_cur_namespace_intrinsic (void);
-extern SLang_Array_Type *_pSLang_apropos (char *, char *, unsigned int);
-extern void _pSLang_implements_intrinsic (char *);
+extern SLFUTURE_CONST char *_pSLang_cur_namespace_intrinsic (void);
+extern SLang_Array_Type *_pSLang_apropos (SLFUTURE_CONST char *, SLFUTURE_CONST char *, unsigned int);
+extern void _pSLang_implements_intrinsic (SLFUTURE_CONST char *);
 extern SLang_Array_Type *_pSLns_list_namespaces (void);
-extern SLang_Name_Type *_pSLns_locate_hashed_name (SLang_NameSpace_Type *, char *, unsigned long);
+extern SLang_Name_Type *_pSLns_locate_hashed_name (SLang_NameSpace_Type *, SLCONST char *, unsigned long);
 extern int _pSLns_add_hashed_name (SLang_NameSpace_Type *, SLang_Name_Type *, unsigned long);
 extern SLang_NameSpace_Type *_pSLns_find_object_namespace (SLang_Name_Type *nt);
-extern SLang_Name_Type *_pSLns_locate_name (SLang_NameSpace_Type *, char *);
-extern SLang_NameSpace_Type *_pSLns_get_private_namespace (char *name, char *nsname);
-extern SLang_NameSpace_Type *_pSLns_create_namespace2 (char *name, char *nsname);
+extern SLang_Name_Type *_pSLns_locate_name (SLang_NameSpace_Type *, SLCONST char *);
+extern SLang_NameSpace_Type *_pSLns_get_private_namespace (SLFUTURE_CONST char *name, SLFUTURE_CONST char *nsname);
+extern SLang_NameSpace_Type *_pSLns_create_namespace2 (SLFUTURE_CONST char *name, SLFUTURE_CONST char *nsname);
 
 extern int _pSLang_Trace;
-extern char *_pSLang_current_function_name (void);
+extern char SLCONST *_pSLang_current_function_name (void);
 
-extern int _pSLang_trace_fun(char *);
+extern int _pSLang_trace_fun(SLFUTURE_CONST char *);
 
 /* This is a bitmapped variable */
-extern int _pSLang_Compile_Line_Num_Info;
+/* extern int _pSLang_Compile_Line_Num_Info; */
 #if SLANG_HAS_BOSEOS
 extern int _pSLang_Compile_BOSEOS;
+extern int _pSLang_Compile_BOFEOF;
 extern int _pSLang_init_boseos (void);
-extern int _pSLcall_bos_handler (char *, int);
+extern int _pSLcall_bos_handler (SLFUTURE_CONST char *, int);
 extern int _pSLcall_eos_handler (void);
-extern int _pSLcall_debug_hook (char *file, int linenum);
+extern int _pSLcall_bof_handler (SLFUTURE_CONST char *, SLFUTURE_CONST char *);
+extern int _pSLcall_eof_handler (void);
+extern int _pSLcall_debug_hook (SLFUTURE_CONST char *file, int linenum);
 /* extern int _pSLcall_debug_hook (char *file, int linenum, char *funct); */
 #endif
 
-extern char *_pSLstring_dup_hashed_string (char *, unsigned long);
-extern unsigned long _pSLcompute_string_hash (char *);
-extern char *_pSLstring_make_hashed_string (char *, unsigned int, unsigned long *);
+extern char *_pSLstring_dup_hashed_string (SLCONST char *, unsigned long);
+extern unsigned long _pSLcompute_string_hash (SLCONST char *);
+extern char *_pSLstring_make_hashed_string (SLCONST char *, unsigned int, unsigned long *);
 extern void _pSLfree_hashed_string (char *, unsigned int, unsigned long);
-unsigned long _pSLstring_hash (unsigned char *, unsigned char *);
+unsigned long _pSLstring_hash (SLCONST unsigned char *, SLCONST unsigned char *);
 extern int _pSLinit_slcomplex (void);
 
 extern int _pSLang_init_slstrops (void);
 extern int _pSLstrops_do_sprintf_n (int);
 extern int _pSLang_sscanf (void);
-extern double _pSLang_atof (char *);
+extern double _pSLang_atof (SLFUTURE_CONST char *);
 extern int _pSLang_init_bstring (void);
 extern int _pSLang_init_sltime (void);
 extern void _pSLpack (void);
@@ -709,7 +729,7 @@ extern int _pSLarith_typecast (SLtype, VOID_STAR, unsigned int,
 			      SLtype, VOID_STAR);
 
 extern int SLang_push(SLang_Object_Type *);
-extern int SLadd_global_variable (char *);
+extern int SLadd_global_variable (SLCONST char *);
 
 extern int _pSLdo_pop (void);
 extern unsigned int _pSLsys_getkey (void);
@@ -722,10 +742,11 @@ extern unsigned int _pSLpc_convert_scancode (unsigned int, unsigned int, int);
 #endif
 
 typedef struct _pSLterminfo_Type SLterminfo_Type;
-extern SLterminfo_Type *_pSLtt_tigetent (char *);
-extern char *_pSLtt_tigetstr (SLterminfo_Type *, char *);
-extern int _pSLtt_tigetnum (SLterminfo_Type *, char *);
-extern int _pSLtt_tigetflag (SLterminfo_Type *, char *);
+extern SLterminfo_Type *_pSLtt_tigetent (SLCONST char *);
+extern void _pSLtt_tifreeent (SLterminfo_Type *);
+extern char *_pSLtt_tigetstr (SLterminfo_Type *, SLCONST char *);
+extern int _pSLtt_tigetnum (SLterminfo_Type *, SLCONST char *);
+extern int _pSLtt_tigetflag (SLterminfo_Type *, SLCONST char *);
 
 #if SLTT_HAS_NON_BCE_SUPPORT
 extern int _pSLtt_get_bce_color_offset (void);
@@ -850,8 +871,8 @@ struct _pSLang_Class_Type
    int (*cl_foreach) (SLtype, SLang_Foreach_Context_Type *);
 
    /* Structure access: get and put (assign to) fields */
-   int (*cl_sput) (SLtype, char *);
-   int (*cl_sget) (SLtype, char *);
+   int (*cl_sput) (SLtype, SLFUTURE_CONST char *);
+   int (*cl_sget) (SLtype, SLFUTURE_CONST char *);
 
    /* File I/O */
    int (*cl_fread) (SLtype, FILE *, VOID_STAR, unsigned int, unsigned int *);
@@ -896,8 +917,8 @@ extern int _pSLclass_add_arith_unary_op (SLtype type,
 					int (*r)(int, SLtype, SLtype *));
 #endif
 
-extern int _pSLclass_get_unary_opcode (char *name);
-extern int _pSLclass_get_binary_opcode (char *name);
+extern int _pSLclass_get_unary_opcode (SLCONST char *name);
+extern int _pSLclass_get_binary_opcode (SLCONST char *name);
 extern int _pSLclass_is_same_obj (SLang_Object_Type *a, SLang_Object_Type *b);
 extern int _pSLclass_obj_eqs (SLang_Object_Type *a, SLang_Object_Type *b);
 
@@ -910,39 +931,45 @@ extern void _pSLang_set_arith_type (SLtype, unsigned char);
 #if SLANG_OPTIMIZE_FOR_SPEED
 extern SLclass_Type _pSLang_get_class_type (SLtype);
 extern void _pSLang_set_class_info (SLtype, SLang_Class_Type *);
+extern int _pSLarray_bin_op (SLang_Object_Type *, SLang_Object_Type *, int);
 #endif
 extern int _pSLarith_bin_op (SLang_Object_Type *, SLang_Object_Type *, int);
 
 extern int _pSLarray_add_bin_op (SLtype);
+extern int _pSLarray_push_elem_ref (void);
+extern int _pSLang_push_array (SLang_Array_Type *, int);   /* NULL not allowed */
 
 typedef struct
 {
-   char *name;
+   SLCONST char *name;
    SLang_Name_Type *next;
    char name_type;
 
    int unary_op;
 }
 SLang_Arith_Unary_Type;
-extern int _pSLadd_arith_unary_table (SLang_Arith_Unary_Type *tbl, char *);
+extern int _pSLadd_arith_unary_table (SLang_Arith_Unary_Type *tbl, SLFUTURE_CONST char *);
 
 typedef struct
 {
-   char *name;
+   SLCONST char *name;
    SLang_Name_Type *next;
    char name_type;
 
    int binary_op;
 }
 SLang_Arith_Binary_Type;
-extern int _pSLadd_arith_binary_table (SLang_Arith_Binary_Type *tbl, char *);
+extern int _pSLadd_arith_binary_table (SLang_Arith_Binary_Type *tbl, SLFUTURE_CONST char *);
+
+extern int _pSLang_do_binary_ab (int op, SLang_Object_Type *obja, SLang_Object_Type *objb);
 
 extern int _pSLang_call_funptr (SLang_Name_Type *);
-extern void _pSLset_double_format (char *);
-extern SLang_Name_Type *_pSLlocate_global_name (char *);
-extern SLang_Name_Type *_pSLlocate_name (char *);
+extern void _pSLset_double_format (SLCONST char *);
+extern SLCONST char *_pSLget_double_format (void);
+extern SLang_Name_Type *_pSLlocate_global_name (SLCONST char *);
+extern SLang_Name_Type *_pSLlocate_name (SLCONST char *);
 
-extern char *_pSLdefines[];
+extern SLFUTURE_CONST char *_pSLdefines[];
 
 #define SL_ERRNO_NOT_IMPLEMENTED	0x7FFF
 extern int _pSLerrno_errno;
@@ -961,6 +988,7 @@ extern int _pSLarray_aput (void);
 extern int _pSLarray_aget (void);
 extern int _pSLarray_aget1 (unsigned int);
 extern int _pSLarray_inline_implicit_array (void);
+extern int _pSLarray_inline_implicit_arrayn (void);
 extern int _pSLarray_inline_array (void);
 extern int _pSLarray_wildcard_array (void);
 
@@ -990,10 +1018,12 @@ SLang_create_array1 (SLtype, int, VOID_STAR, SLindex_Type *, unsigned int, int);
 
 extern int _pSLassoc_aput (SLtype, unsigned int);
 extern int _pSLassoc_aget (SLtype, unsigned int);
+extern int _pSLassoc_inc_value (unsigned int, int val);
 
 extern int _pSLcompile_push_context (SLang_Load_Type *);
 extern int _pSLcompile_pop_context (void);
 extern int _pSLang_Auto_Declare_Globals;
+extern int _pSLang_Load_File_Verbose;
 
 typedef struct _pSLang_Token_Type
 {
@@ -1012,7 +1042,7 @@ typedef struct _pSLang_Token_Type
 	 * We can avoid the issues associated with this by just storing 
 	 * floating point values as strings.
 	 */
-	char *s_val;		       /* Used for IDENT_TOKEN, DOUBLE_TOKEN, etc...  */
+	SLFUTURE_CONST char *s_val;		       /* Used for IDENT_TOKEN, DOUBLE_TOKEN, etc...  */
 	
 	SLang_BString_Type *b_val;
      } v;
@@ -1031,10 +1061,10 @@ extern void _pSLcompile (_pSLang_Token_Type *);
 extern void (*_pSLcompile_ptr)(_pSLang_Token_Type *);
 
 /* slmisc.c */
-extern char *_pSLskip_whitespace (char *s);
+extern char *_pSLskip_whitespace (SLCONST char *s);
 
 /* slospath.c */
-extern char *_pSLpath_find_file (char *);   /* slstring returned */
+extern char *_pSLpath_find_file (SLFUTURE_CONST char *, int);   /* slstring returned */
 
 /* Read but do not set this variable. */
 extern volatile int _pSLang_Error;
@@ -1066,6 +1096,7 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define NOP_TOKEN	0x05
 #define FARG_TOKEN	0x06
 #define TMP_TOKEN	0x07
+#define QUALIFIER_TOKEN	0x08
 
 #define RESERVED1_TOKEN	0x0A	       /* \n */
 #define RESERVED2_TOKEN	0x0D	       /* \r */
@@ -1119,20 +1150,24 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define NAMESPACE_TOKEN	0x34
 
 /* Operators */
-#define POW_TOKEN	0x38
-
 /* The order here must match the order in the Binop_Level table in slparse.c */
-#define FIRST_BINARY_OP	0x39
+#define FIRST_BINARY_OP	0x36
+#define SC_AND_TOKEN	0x36
+#define SC_OR_TOKEN	0x37
+#define POW_TOKEN	0x38
 #define ADD_TOKEN	0x39
 #define SUB_TOKEN	0x3a
 #define TIMES_TOKEN	0x3b
 #define DIV_TOKEN	0x3c
+
 #define LT_TOKEN	0x3d
 #define LE_TOKEN	0x3e
 #define GT_TOKEN	0x3f
 #define GE_TOKEN	0x40
 #define EQ_TOKEN	0x41
 #define NE_TOKEN	0x42
+#define IS_COMPARE_OP(x) (((x) >= LT_TOKEN) && ((x) <= NE_TOKEN))
+
 #define AND_TOKEN	0x43
 #define OR_TOKEN	0x44
 #define MOD_TOKEN	0x45
@@ -1142,7 +1177,6 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define BXOR_TOKEN	0x49
 #define BOR_TOKEN	0x4a
 #define POUND_TOKEN	0x4b	       /* matrix multiplication */
-
 #define LAST_BINARY_OP	 0x4b
 #define IS_BINARY_OP(t)	 ((t >= FIRST_BINARY_OP) && (t <= LAST_BINARY_OP))
 
@@ -1179,7 +1213,6 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define POST_MINUSMINUS_TOKEN	0x61
 
 /* Directives */
-#define FIRST_DIRECTIVE_TOKEN	0x62
 #define IFNOT_TOKEN	0x62
 #define IF_TOKEN	0x63
 #define ELSE_TOKEN	0x64
@@ -1226,6 +1259,10 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define THROW_TOKEN	0x89
 #define FINALLY_TOKEN	0x8a
 
+#define THEN_TOKEN	0x8d
+
+#define STRUCT_WITH_ASSIGN_TOKEN	0x8e
+
 /* Note: the order here must match the order of the generic assignment tokens.
  * Also, the first token of each group must be the ?_ASSIGN_TOKEN.
  * slparse.c exploits this order, as well as slang.h.
@@ -1282,11 +1319,23 @@ extern SLuchar_Type *_pSLinterp_encode_wchar (SLwchar_Type wch,
 #define LAST_ASSIGN_TOKEN		0xCA
 #define IS_ASSIGN_TOKEN(t) (((t)>=FIRST_ASSIGN_TOKEN)&&((t)<=LAST_ASSIGN_TOKEN))
 
+#define _DEREF_FUNCALL_TOKEN		0xCE
+
+#define LOOP_THEN_TOKEN			0xD0
+/* #define LOOP_ELSE_TOKEN			0xD1 */
+
+#define _COMPARE_TOKEN			0xD8   /* (a < b < c ...) */
+#define _ARRAY_ELEM_REF_TOKEN		0xD9   /* &X[i] */
+#define _STRUCT_FIELD_REF_TOKEN		0xDA   /* &S.a */
+
 #define _INLINE_ARRAY_TOKEN		0xE0
-#define _INLINE_IMPLICIT_ARRAY_TOKEN	0xE1
+#define _INLINE_IMPLICIT_ARRAY_TOKEN	0xE1   /* [a:b:c] */
 #define _NULL_TOKEN			0xE2
 #define _INLINE_WILDCARD_ARRAY_TOKEN	0xE3
 #define _INLINE_LIST_TOKEN		0xE4
+#define _INLINE_IMPLICIT_ARRAYN_TOKEN	0xE5   /* [a:b:#n] */
+
+#define ESC_STRING_DOLLAR_TOKEN		0xF0
 
 #define BOS_TOKEN			0xFA
 #define EOS_TOKEN			0xFB
@@ -1320,15 +1369,15 @@ typedef struct
 _pSLEval_Context;
 
 extern int _pSLget_token (_pSLang_Token_Type *);
-extern void _pSLparse_error (int, char *, _pSLang_Token_Type *, int);
+extern void _pSLparse_error (int, SLCONST char *, _pSLang_Token_Type *, int);
 extern void _pSLparse_start (SLang_Load_Type *);
 extern int _pSLget_rpn_token (_pSLang_Token_Type *);
 extern void _pSLcompile_byte_compiled (void);
 
-extern int (*_pSLprep_eval_hook) (char *);
+extern int (*_pSLprep_eval_hook) (SLFUTURE_CONST char *);
 
 extern int _pSLsecure_issetugid (void);
-extern char *_pSLsecure_getenv (char *);
+extern char *_pSLsecure_getenv (SLCONST char *);
 
 /* Error Handling */
 extern int _pSLang_init_exceptions (void);
@@ -1336,32 +1385,41 @@ extern int _pSLerr_init (void);
 extern void _pSLerr_deinit (void);
 extern int _pSLerr_suspend_messages (void);
 extern int _pSLerr_resume_messages (void);
-extern int _pSLerr_traceback_msg (char *, ...) SLATTRIBUTE_PRINTF(1,2);
-extern void _pSLerr_dump_msg (char *, ...) SLATTRIBUTE_PRINTF(1,2);
-extern void _pSLerr_clear_error (void);
+extern int _pSLerr_traceback_msg (SLFUTURE_CONST char *, ...) SLATTRIBUTE_PRINTF(1,2);
+extern void _pSLerr_dump_msg (SLFUTURE_CONST char *, ...) SLATTRIBUTE_PRINTF(1,2);
+extern void _pSLerr_clear_error (int);
+extern void _pSLang_verror (int, SLCONST char *, ...) SLATTRIBUTE_PRINTF(2,3);
+extern int _pSLsnprintf (char *, unsigned int, SLFUTURE_CONST char *, ...) SLATTRIBUTE_PRINTF(3,4);
+
+typedef struct _pSLerr_Error_Queue_Type _pSLerr_Error_Queue_Type;
+extern _pSLerr_Error_Queue_Type *_pSLerr_new_error_queue (int);
+extern int _pSLerr_set_error_queue (_pSLerr_Error_Queue_Type *);
+extern void _pSLerr_delete_error_queue (_pSLerr_Error_Queue_Type *);
+
 extern void _pSLerr_print_message_queue (void);
 
-extern int _pSLerr_set_line_info (char *, int, char *);
-extern int _pSLang_pop_error_context (void);
+extern int _pSLerr_set_line_info (SLFUTURE_CONST char *, int, SLFUTURE_CONST char *);
+extern int _pSLang_pop_error_context (int);
 extern int _pSLang_push_error_context (void);
 extern void (*_pSLinterpreter_Error_Hook)(int);
 extern int _pSLerr_init_interp_exceptions (void);
 extern int _pSLerr_get_last_error (void);
 extern int _pSLerr_pop_exception (int *);
 extern void _pSLerr_free_queued_messages (void);
-extern char *_pSLerr_get_error_from_queue (void);
+
+extern char *_pSLerr_get_error_from_queue (_pSLerr_Error_Queue_Type *);
 extern int _pSLerr_throw (void);
-extern int (*_pSLerr_New_Exception_Hook)(char *name, char *desc, int error_code);
+extern int (*_pSLerr_New_Exception_Hook)(SLFUTURE_CONST char *name, SLFUTURE_CONST char *desc, int error_code);
 
 #if SLANG_HAS_DEBUGGER_SUPPORT
 typedef struct
 {
    char **locals;
    unsigned int nlocals;
-   char *file;
+   SLCONST char *file;
    unsigned int line;
-   char *function;
-   char *ns;
+   SLCONST char *function;
+   SLCONST char *ns;
 }
 _pSLang_Frame_Info_Type;
 
@@ -1374,6 +1432,9 @@ extern int _pSLang_get_frame_depth (void);
 
 #if SLANG_HAS_FLOAT
 extern int _pSLmath_isnan (double x);
+extern int _pSLmath_isinf (double x);
+extern double _pSLang_NaN;
+extern double _pSLang_Inf;
 #endif
 
 #if SLANG_HAS_SIGNALS
